@@ -825,6 +825,10 @@ static __isl_give isl_ast_build *update_values(
  * Otherwise, we would indirectly intersect the build domain with itself,
  * which can lead to inefficiencies, in particular if the build domain
  * contains any unknown divs.
+ *
+ * The pending and generated sets are not updated by this function to
+ * match the updated domain.
+ * The caller still needs to call isl_ast_build_set_pending_generated.
  */
 __isl_give isl_ast_build *isl_ast_build_set_loop_bounds(
 	__isl_take isl_ast_build *build, __isl_take isl_basic_set *bounds)
@@ -835,12 +839,10 @@ __isl_give isl_ast_build *isl_ast_build_set_loop_bounds(
 	if (!build)
 		goto error;
 
-	bounds = isl_basic_set_preimage_multi_aff(bounds,
-					isl_multi_aff_copy(build->values));
 	build = update_values(build, isl_basic_set_copy(bounds));
 	if (!build)
 		goto error;
-	set = isl_set_from_basic_set(isl_basic_set_copy(bounds));
+	set = isl_set_from_basic_set(bounds);
 	if (isl_ast_build_has_affine_value(build, build->depth)) {
 		set = isl_set_eliminate(set, isl_dim_set, build->depth, 1);
 		set = isl_set_compute_divs(set);
@@ -848,24 +850,11 @@ __isl_give isl_ast_build *isl_ast_build_set_loop_bounds(
 							isl_set_copy(set));
 		build->domain = isl_set_intersect(build->domain, set);
 	} else {
-		isl_basic_set *generated, *pending;
-
-		pending = isl_basic_set_copy(bounds);
-		pending = isl_basic_set_drop_constraints_involving_dims(pending,
-					isl_dim_set, build->depth, 1);
-		build->pending = isl_set_intersect(build->pending,
-					isl_set_from_basic_set(pending));
-		generated = isl_basic_set_copy(bounds);
-		generated = isl_basic_set_drop_constraints_not_involving_dims(
-				    generated, isl_dim_set, build->depth, 1);
-		build->generated = isl_set_intersect(build->generated,
-					isl_set_from_basic_set(generated));
 		build->domain = isl_set_intersect(build->domain, set);
 		build = isl_ast_build_include_stride(build);
 		if (!build)
 			goto error;
 	}
-	isl_basic_set_free(bounds);
 
 	if (!build->domain || !build->pending || !build->generated)
 		return isl_ast_build_free(build);
@@ -877,6 +866,49 @@ error:
 	return NULL;
 }
 
+/* Update the pending and generated sets of "build" according to "bounds".
+ * If the build has an affine value at the current depth,
+ * then isl_ast_build_set_loop_bounds has already set the pending set.
+ * Otherwise, do it here.
+ */
+__isl_give isl_ast_build *isl_ast_build_set_pending_generated(
+	__isl_take isl_ast_build *build, __isl_take isl_basic_set *bounds)
+{
+	isl_basic_set *generated, *pending;
+
+	if (!build)
+		goto error;
+
+	if (isl_ast_build_has_affine_value(build, build->depth)) {
+		isl_basic_set_free(bounds);
+		return build;
+	}
+
+	build = isl_ast_build_cow(build);
+	if (!build)
+		goto error;
+
+	pending = isl_basic_set_copy(bounds);
+	pending = isl_basic_set_drop_constraints_involving_dims(pending,
+				isl_dim_set, build->depth, 1);
+	build->pending = isl_set_intersect(build->pending,
+				isl_set_from_basic_set(pending));
+	generated = bounds;
+	generated = isl_basic_set_drop_constraints_not_involving_dims(
+			    generated, isl_dim_set, build->depth, 1);
+	build->generated = isl_set_intersect(build->generated,
+				isl_set_from_basic_set(generated));
+
+	if (!build->pending || !build->generated)
+		return isl_ast_build_free(build);
+
+	return build;
+error:
+	isl_ast_build_free(build);
+	isl_basic_set_free(bounds);
+	return NULL;
+}
+
 /* Intersect build->domain with "set", where "set" is specified
  * in terms of the internal schedule domain.
  */
@@ -950,31 +982,6 @@ __isl_give isl_ast_build *isl_ast_build_replace_pending_by_guard(
 	return build;
 }
 
-/* Intersect build->pending and build->domain with "set",
- * where "set" is specified in terms of the internal schedule domain.
- */
-__isl_give isl_ast_build *isl_ast_build_restrict_pending(
-	__isl_take isl_ast_build *build, __isl_take isl_set *set)
-{
-	set = isl_set_compute_divs(set);
-	build = isl_ast_build_restrict_internal(build, isl_set_copy(set));
-	build = isl_ast_build_cow(build);
-	if (!build)
-		goto error;
-
-	build->pending = isl_set_intersect(build->pending, set);
-	build->pending = isl_set_coalesce(build->pending);
-
-	if (!build->pending)
-		return isl_ast_build_free(build);
-
-	return build;
-error:
-	isl_ast_build_free(build);
-	isl_set_free(set);
-	return NULL;
-}
-
 /* Intersect build->domain with "set", where "set" is specified
  * in terms of the external schedule domain.
  */
@@ -2193,6 +2200,18 @@ __isl_give isl_set *isl_ast_build_specialize(__isl_keep isl_ast_build *build,
 					isl_multi_aff_copy(build->values));
 }
 
+/* Plug in the known affine values of outer loop iterators in "bset".
+ */
+__isl_give isl_basic_set *isl_ast_build_specialize_basic_set(
+	__isl_keep isl_ast_build *build, __isl_take isl_basic_set *bset)
+{
+	if (!build)
+		return isl_basic_set_free(bset);
+
+	return isl_basic_set_preimage_multi_aff(bset,
+					isl_multi_aff_copy(build->values));
+}
+
 /* Simplify the map "map" based on what we know about
  * the iterators of already generated loops.
  *
